# SPDX-FileCopyrightText: 2020-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/_scripts/bundlerun.py
# DESCRIPTION:    Script to run bundle of services
# CREATED:        6.12.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Command-line script to run a bundle of Saturnin services.

This script provides a convenient way to execute a collection of Saturnin
services as a coordinated bundle. It parses command-line arguments for
configuration files and logging levels, then uses
`saturnin.component.bundle.BundleExecutor` to manage the lifecycle
of all services within the bundle.
"""

from __future__ import annotations

import logging
from argparse import Action, ArgumentDefaultsHelpFormatter, ArgumentParser
from configparser import ConfigParser
from logging.config import fileConfig

from saturnin.base import SECTION_BUNDLE, directory_scheme
from saturnin.component.bundle import BundleExecutor
from saturnin.component.controller import Outcome

from firebird.base.config import EnvExtendedInterpolation
from firebird.base.logging import get_logger
from firebird.base.trace import trace_manager

LOG_FORMAT = '%(levelname)s [%(processName)s/%(threadName)s] %(message)s'

class UpperAction(Action):
    """Converts argument to uppercase.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values.upper())

def main(description: str | None=None, bundle_config: str | None=None):
    """Saturnin script to run a bundle of services.

    This function sets up command-line argument parsing, initializes logging
    and tracing, and then uses `BundleExecutor` to run the specified
    service bundle based on the provided configuration.

    Arguments:
      description:   Custom description for the script, shown when
                     `--help` is used. Defaults to a standard description.
      bundle_config: Default path for the `BUNDLE-CONFIG` argument.
                     If `None`, `BUNDLE-CONFIG` becomes a required positional argument.

    Command-line usage (auto-generated by argparse)::

      saturnin-bundle [-h] [-c CONFIG] [-s SECTION] [-q] [-o]
                      [-l {critical,fatal,error,warn,warning,info,debug,notset}]
                      BUNDLE-CONFIG

      Positional arguments:
        BUNDLE-CONFIG         Path to service bundle configuration file.

      Optional arguments:
        -h, --help            show this help message and exit
        -c CONFIG, --config CONFIG
                              Path to additional configuration file. Could be specified
                              multiple times. (default: None)
        -s SECTION, --section SECTION
                              Configuration section name (default: bundle)
        -q, --quiet           Suppress console output. (default: False)
        -o, --outcome         Always print service execution outcome (default: False)
        -l {critical,fatal,error,warn,warning,info,debug,notset},
        --log-level {critical,fatal,error,warn,warning,info,debug,notset}
                              Logging level (default: None)
    """
    if description is None:
        description = "Saturnin script to run bundle of services."
    parser: ArgumentParser = ArgumentParser(description=description,
                                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('service', metavar='BUNDLE-CONFIG',
                        help="Path to service bundle configuration file.",
                        nargs=1 if bundle_config is None else '?',
                        default=bundle_config)
    parser.add_argument('-c','--config', metavar='CONFIG', action='append',
                        help="Path to additional configuration file. "
                             "Could be specified multiple times.")
    parser.add_argument('-s', '--section', help="Configuration section name",
                        default=SECTION_BUNDLE)
    parser.add_argument('-q', '--quiet', action='store_true', help="Suppress console output.",
                        default=False)
    parser.add_argument('-o','--outcome', action='store_true',
                        help="Always print service execution outcome", default=False)
    parser.add_argument('-l', '--log-level', action=UpperAction,
                        choices=[x.lower() for x in logging._nameToLevel
                                 if isinstance(x, str)],
                        help="Logging level")

    args = parser.parse_args()

    main_config: ConfigParser = ConfigParser(interpolation=EnvExtendedInterpolation())
    cfg_files = [str(directory_scheme.logging_conf)]
    if args.config:
        cfg_files.extend(args.config)
    if isinstance(args.service, list):
        cfg_files.extend(args.service)
    else:
        cfg_files.append(args.service)
    cfg_files = main_config.read(cfg_files)
    # Logging configuration
    if main_config.has_section('loggers'):
        fileConfig(main_config)
    else:
        logging.basicConfig(format=LOG_FORMAT)
    log: logging.Logger = get_logger('saturnin')
    if args.log_level is not None:
        log.setLevel(args.log_level)
    # trace configuration
    if main_config.has_section('trace'):
        trace_manager.load_config(main_config)

    try:
        with BundleExecutor() as executor:
            executor.configure(cfg_files, section=args.section)
            result = executor.run()
            if not args.quiet:
                for name, outcome, details in result:
                    if outcome is not Outcome.OK or args.outcome:
                        print(f'{name}: {outcome.value}')
                        if details:
                            for line in details:
                                print(f' {line}')
    except Exception as exc:
        log.exception("Service execution failed")
        parser.exit(1, f'{exc!s}\n')
    finally:
        logging.shutdown()

if __name__ == '__main__':
    main()
